#![feature(min_const_generics)]
#![feature(let_chains)]
#[macro_use]
extern crate criterion;
use criterion::{BenchmarkId, Criterion};
use rand::prelude::*;
use rand::{thread_rng, Rng};
use Xorfilter::*;
pub fn xorfilter_new(c: &mut Criterion) {
    //const size: usize = 300000;
    const size: usize = 10;
    let mut group = c.benchmark_group("xorfilter_new");
    let group = group.sample_size(size);

    let keys = create_keys::<size>();

    group.bench_with_input(BenchmarkId::new("xorfilter_new", size), &keys, |b, keys| {
        b.iter(|| XorFilter::<u8, size>::new(*keys));
    });
}

pub fn xorfilter_contains(c: &mut Criterion) {
    //const size: usize = 300000;
    const size: usize = 10;
    let mut group = c.benchmark_group("xorfilter_contains");
    let group = group.sample_size(size);

    let mut keys = [0; size];
    let mut rng = thread_rng();
    for i in 0..size {
        keys[i] = (rng.gen::<u64>());
    }

    let x: XorFilter<u8, size> = XorFilter::new(keys);

    group.bench_function(BenchmarkId::new("xorfilter_contains", size), |b| {
        let key: u64 = rng.gen();
        b.iter(|| x.contains(key));
    });
}

fn create_keys<const N: usize>() -> [u64; N] {
    //use rand::prelude::*;
    //use rand::{thread_rng, Rng};
    let mut rng = thread_rng();
    let mut keys = [0; N];
    for i in 0..N {
        keys[i] = (rng.gen::<u64>());
    }

    keys
}

criterion_group!(xorfilter, xorfilter_new, xorfilter_contains);
criterion_main!(xorfilter);
